# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine (SIP Examine)
# Matches SIP dialogs by time, regex or pcap packet filter; returns concise SIP
# packet info.
# Copyright (C) 2011-2015,2018 Walter Doekes, OSSO B.V.

__author__ = 'Walter Doekes'
__copyright__ = 'Copyright (C) Walter Doekes, OSSO B.V. 2011-2015,2018'
__licence__ = 'GPLv3+'
__version__ = '0.2_dev'



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine Base Protocol lib
# Copyright (C) 2011-2015 Walter Doekes, OSSO B.V.

import datetime


class IpPacket(object):
    '''
    IpPacket holds a TCP or UDP packet with data. Call IpPacket.create()
    when you're creating a new packet from data. It will search the list
    of registered refined types so you can get a subtype with more
    advanced methods instead.
    '''
    __types = []  # the list of registered subtypes

    @classmethod
    def create(cls, datetime, ip_proto, from_, to, data):
        '''
        Create a new IpPacket or, if possible, a refined subtype. Never
        call __init__ directly. Overriding this in your subclass doesn't
        make sense. This will be called anyway.
        '''
        # Create an IpPacket that will be returned if no one wants to
        # take it.
        default_packet = cls(datetime, ip_proto, from_, to, data)

        probabilities = []
        for type in cls.__types:
            probability = type.type_probability(default_packet)
            if probability != 0.0:
                probabilities.append((probability, type))

        if probabilities:
            probabilities.sort(key=lambda x: (-x[0], x[1].__name__))
            return probabilities[0][1](datetime, ip_proto, from_, to, data)

        return default_packet

    @classmethod
    def register_subtype(cls, class_):
        '''
        Register a new IpPacket subtype. Only subclasses of IpPacket may
        be registered.
        '''
        if not issubclass(class_, cls):
            raise TypeError('Not an IpPacket', class_)

        try:
            random_packet = IpPacket(datetime.datetime.now(), 'TCP',
                                     ('1.2.3.4', 1234), ('1.2.3.4', 1234), '')
            probability_check = class_.type_probability(random_packet)
            if not (0.0 <= probability_check <= 1.0):
                raise NotImplementedError()
        except NotImplementedError:
            raise TypeError('IpPacket subtype does not implement '
                            'type_probability() correctly: Expected a float '
                            'between 0 and 1')

        cls.__types.append(class_)

    @classmethod
    def type_probability(cls, datetime, ip_proto, from_, to, data):
        '''
        Return a number between 0.0 and 1.0 stating the probability that
        the packet is your type. We need to trust you on this. Don't let
        us down.
        '''
        raise NotImplementedError()

    def __init__(self, datetime, ip_proto, from_, to, data):
        self.datetime = datetime
        self.ip_proto = ip_proto
        self.from_ = from_
        self.to = to
        self.data = data

    def __repr__(self):
        summary = '(null)'
        if self.data:
            summary = self.data[0:12] + '...'
        return '<%s(%s, %s, %s, %s, %d B %r)>' % (
            self.__class__.__name__,
            self.datetime.strftime('%y%m%d:%H%M%S.%f'),
            self.ip_proto,
            self.from_,
            self.to,
            len(self.data),
            summary
        )


# # Simple test/example
# now = datetime.datetime.now()
# data = 'A bit of data'
# ip_packet = IpPacket.create(
#     now,
#     'TCP',
#     ('1.2.3.4', 1234),
#     ('1.2.3.4', 1234),
#     'A bit of data'
# )
# assert isinstance(ip_packet, IpPacket), \
#     'Packet is of type: %r' % (type(ip_packet),)
# assert ip_packet.datetime == now
# assert ip_packet.data == data



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine Argparse wrapper
# Copyright (C) 2015 Walter Doekes, OSSO B.V.

try:
    from argparse import ArgumentParser
except ImportError:
    from .argparse_1_2_1 import ArgumentParser


class ArgumentParser14191(ArgumentParser):
    """ArgumentParser from argparse that handles out-of-order positional
    arguments.

    This is a workaround created by Glenn Linderman in July 2012. You
    can now do this:

        parser = ArgumentParser14191()
        parser.add_argument('-f', '--foo')
        parser.add_argument('cmd')
        parser.add_argument('rest', nargs='*')
        # some of these would fail with the regular parser:
        for args, res in (('-f1 cmd 1 2 3', 'ok'),
                          ('cmd -f1 1 2 3', 'would_fail'),
                          ('cmd 1 -f1 2 3', 'would_fail'),
                          ('cmd 1 2 3 -f1', 'ok')):
            try: out = parser.parse_args(args.split())
            except: print('args', 'failed', res)
            # out: Namespace(cmd='cmd', foo='1', rest=['1', '2', '3'])

    Bugs: http://bugs.python.org/issue14191
    Files: http://bugs.python.org/file26273/t18a.py
    Changes: renamed to ArgumentParser14191 ** PEP cleaned ** hidden
      ErrorParser inside ArgumentParser14191 ** documented ** used
      new-style classes super calls  (Walter Doekes, March 2015)
    """
    class ErrorParser(ArgumentParser):
        def __init__(self, *args, **kwargs):
            self.__errorobj = None
            super(ArgumentParser14191.ErrorParser, self).__init__(
                *args, add_help=False, **kwargs)

        def error(self, message):
            if self.__errorobj:
                self.__errorobj.error(message)
            else:
                ArgumentParser.error(self, message)

        def seterror(self, errorobj):
            self.__errorobj = errorobj

    def __init__(self, *args, **kwargs):
        self.__setup = False
        self.__opt = ArgumentParser14191.ErrorParser(*args, **kwargs)
        super(ArgumentParser14191, self).__init__(*args, **kwargs)
        self.__opt.seterror(self)
        self.__setup = True

    def add_argument(self, *args, **kwargs):
        super(ArgumentParser14191, self).add_argument(*args, **kwargs)
        if self.__setup:
            chars = self.prefix_chars
            if args and len(args[0]) and args[0][0] in chars:
                self.__opt.add_argument(*args, **kwargs)

    def parse_args(self, args=None, namespace=None):
        ns, remain = self.__opt.parse_known_args(args, namespace)
        ns = super(ArgumentParser14191, self).parse_args(remain, ns)
        return ns



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine File reader lib
# Copyright (C) 2018 Walter Doekes, OSSO B.V.


def extract_to_tempfile(filename):
    has_tempfile = False

    if filename.endswith('.gz'):
        import gzip
        import tempfile

        gunzipped = tempfile.NamedTemporaryFile(delete=False)
        with gzip.open(filename, 'rb') as fp:
            gunzipped.write(fp.read())
        gunzipped.flush()
        has_tempfile = True
        filename = gunzipped.name

    return filename, has_tempfile



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine main (SIP Examine)
# Copyright (C) 2011-2015,2018 Walter Doekes, OSSO B.V.

import re


def dateskew_filter(reader, skew):
    """
    Alter reader to shift all iterated object dates by dateskew.
    """
    for packet in reader:
        packet.datetime += skew
        yield packet


def minduration_filter(reader, min_duration):
    """
    Filter dialogs by minimum duration.
    """
    for dialog in reader:
        duration = dialog[-1].datetime - dialog[0].datetime
        if duration >= min_duration:
            yield dialog


def maxduration_filter(reader, max_duration):
    """
    Filter dialogs by maximum duration.
    """
    for dialog in reader:
        duration = dialog[-1].datetime - dialog[0].datetime
        if duration <= max_duration:
            yield dialog


def allheaders_filter(reader, header_match):
    """
    Filter dialogs by a regexp which must match all packets in the dialog.
    """
    for dialog in reader:
        for packet in dialog:
            if not packet.search(header_match):
                break
        else:
            yield dialog


def anyheader_filter(reader, header_match):
    """
    Filter dialogs by a regexp which must match any packet in the dialog.
    """
    for dialog in reader:
        for packet in dialog:
            if packet.search(header_match):
                yield dialog
                break


def retransmits_filter(reader, count):
    """
    Filter dialogs where at least count retransmits are observed.
    """
    assert count >= 1
    for dialog in reader:
        dupes = {}
        for packet in dialog:
            if packet.data not in dupes:
                dupes[packet.data] = 0
            else:
                dupes[packet.data] += 1
                if dupes[packet.data] >= count:
                    yield dialog
                    break


def print_dialog(dialog, packet_highlights=None, show_contents=False):
    packet_highlights = packet_highlights or ()  # make sure it's iterable
    if show_contents:
        data_munge = re.compile('^', re.MULTILINE)

    print('[ %s ]' % (dialog[0].callid,))
    for packet in dialog:
        highlights = []
        for i, packet_highlight in enumerate(packet_highlights):
            found_here = packet.search(packet_highlight)
            if found_here:
                if len(packet_highlights) == 1:
                    arrow = '<--'
                else:
                    arrow = '<-%s-' % (chr(ord('a') + i),)
                if not found_here.groups():
                    highlights.append(arrow)
                else:
                    highlights.append('%s %s' %
                                      (arrow, found_here.groups()[0]))

        print(('%s %s:%d > %s:%d %s %s %s' % (
            packet.datetime, packet.from_[0], packet.from_[1],
            packet.to[0], packet.to[1],
            packet.cseq[0], packet.method_and_status, ' '.join(highlights)
        )).rstrip())

        if show_contents:
            print(data_munge.sub('  ', packet.data))

    print('')


def main(reader, packet_matches=None, packet_highlights=None,
         min_duration=None, max_duration=None, show_contents=False):
    # Filter the dialogs
    matching_dialogs = []
    for dialog in reader:
        # print_dialog(dialog, packet_highlights, show_contents=show_contents)
        matching_dialogs.append(dialog)

    # Order dialogs by begin-time and first then print them
    matching_dialogs.sort(key=lambda x: x[0].datetime)
    for dialog in matching_dialogs:
        print_dialog(dialog, packet_highlights, show_contents=show_contents)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine SIP Protocol lib
# Copyright (C) 2011-2015,2018 Walter Doekes, OSSO B.V.

import sys

from collections import defaultdict

from .libproto import IpPacket


class SipPacket(IpPacket):
    @classmethod
    def type_probability(cls, packet):
        # TODO: up the probability if one of more of the ports are 5060
        if packet.data:
            word = packet.data.split(b' ', 1)[0]
            if word in (
                    b'INVITE', b'ACK', b'BYE', b'CANCEL', b'NOTIFY',
                    b'OPTIONS', b'PUBLISH', b'REFER', b'REGISTER',
                    b'SUBSCRIBE', b'UPDATE', b'INFO', b'SIP/2.0'):
                    # XXX: not exhaustive..
                return 0.8
        return 0.0

    # Aliases: f:From, t:To, v:Via, s:Subject, l:Content-Length
    def __init__(self, datetime, ip_proto, from_, to, data):
        super(SipPacket, self).__init__(
            datetime, ip_proto, from_, to, data)
        # FIXME: split up data in header and data
        try:
            data = data.decode('utf-8')  # should be valid utf-8
        except UnicodeDecodeError as err:
            where = (
                b'...' + data[max(0, err.start - 20):err.start] +
                b'<here->' + data[err.start:(err.start + 20)] + b'...')
            sys.stderr.write('(utf-8 decode error near {!r})\n'.format(where))
            data = data.decode('utf-8', 'replace')
        self.headers = data.split('\n')

    @property
    def method(self):
        if not hasattr(self, '_method'):
            word = self.headers[0].split(' ', 1)[0]
            if word != 'SIP/2.0':
                self._method = word
            else:
                self._method = self.cseq[1]
        return self._method

    @property
    def code(self):
        if not hasattr(self, '_code'):
            words = self.headers[0].split(' ', 2)
            if words[0] == 'SIP/2.0':
                self._code = int(words[1])
            else:
                self._code = None
        return self._code

    @property
    def method_and_status(self):
        if self.code:
            return '%s(%s)' % (self.method, self.code)
        return self.method

    @property
    def callid(self):
        if not hasattr(self, '_callid'):
            self._callid = self.get_header('Call-ID', 'i')
        return self._callid

    @property
    def cseq(self):
        if not hasattr(self, '_cseq'):
            data = self.get_header('CSeq')
            if data:
                self._cseq = tuple(i.strip() for i in data.split(None, 1))
            else:
                self._cseq = (None, None)
        return self._cseq

    def get_header(self, header, alt=None):
        # FIXME: doesn't take line-folding into account!
        header = header.lower()
        if alt:
            alt = alt.lower()
        for line in self.headers[1:]:
            # Empty line? done with headers
            if line.strip() == '':
                break
            # Split by colon and match header
            try:
                word, rest = line.split(':', 1)
            except ValueError:
                # Headers broken up too early? Can happen if we're
                # reading an UDP fragment or a TCP stream.. FIXME
                return None
            else:
                word = word.strip().lower()
                if word == header or (alt and word == alt):
                    return rest.strip()
        return None

    def search(self, re_grep):
        for line in self.headers:
            m = re_grep.search(line)
            if m:
                return m
        # FIXME: if body is not in headers, then we want to check body too
        return None


class SipDialog(list):
    '''
    Container for a list of packets belonging to the same dialog.
    '''
    def append(self, item):
        super(SipDialog, self).append(item)
        if hasattr(self, '_is_established'):
            del self._is_established

    def is_established(self):
        if not hasattr(self, '_is_established'):
            cseqs_200 = []
            for i in self[1:]:
                if i.method == 'INVITE' and i.code == 200:
                    cseqs_200.append(i.cseq[0])
                elif i.method == 'ACK' and i.cseq[0] in cseqs_200:
                    self._is_established = True
                    break
            else:
                self._is_established = False
        return self._is_established


class SipDialogs(object):
    '''
    Read from a packet generator and yield SIP dialogs.
    '''
    def __init__(self, packet_generator):
        self.input = packet_generator
        self.dialogs = defaultdict(SipDialog)
        self.latest_datetime = None
        self.yieldable = []

    def __iter__(self):
        self.input = iter(self.input)
        return self

    def __next__(self):
        # Is there anything left to yield
        if self.yieldable:
            return self.yieldable.pop(0)

        # Are we done?
        if not self.input:
            raise StopIteration()

        # Fetch more packets
        try:
            while True:
                packet = self.input.next()
                if isinstance(packet, SipPacket):
                    self.dialogs[packet.callid].append(packet)

                # Check if there is anything we can yield already based
                # on the latest timestamp (dialogs that are old enough
                # to end).
                # FIXME: sometimes we want this in a single dialog
                # (subscribes?)
                if (not self.latest_datetime or
                        ((packet.datetime - self.latest_datetime).seconds >
                         300)):
                    self.update_yieldable(packet.datetime)
                    self.latest_datetime = packet.datetime
                    if self.yieldable:
                        return self.yieldable.pop(0)

        except StopIteration:
            # Time to yield everything we have
            self.yieldable = self.dialogs.values()
            self.yieldable = sorted(
                self.yieldable, key=(lambda x: x[0].datetime))
            self.dialogs.clear()
            # self.input.close()
            self.input = None

        # Return the yieldables or raise StopIteration
        return self.__next__()
    next = __next__

    def update_yieldable(self, latest_datetime):
        # Loop over dialogs, all non-INVITEs shan't be more than 120
        # seconds old. INVITEs may be older, but only if established.
        for k, v in list(self.dialogs.items()):
            yield_it = False
            if (latest_datetime - v[-1].datetime).seconds > 120:
                # FIXME: subscribe establishes a long standing dialog as well
                if v[0].method != 'INVITE':
                    yield_it = True

                elif len(v) == 1 or v[-1].code:
                    # An INVITE without ACK as last packet.
                    yield_it = True

                elif (not v.is_established() and v[-1].method == 'ACK' and
                      v[-2].code >= 300):
                    # An INVITE dialog that failed and was ACKed.
                    yield_it = True

                elif (v.is_established() and v[-1].method == 'BYE' and
                      v[-1].code == 200):
                    # An INVITE dialog that has ended.
                    yield_it = True

                # Yield it?
                if yield_it:
                    self.yieldable.append(v)
                    del self.dialogs[k]

        # Prefer output to be sorted.. it won't be 100%, but it looks
        # better than nothing at all. If you want 100% sorting, I'm
        # afraid you'll have to do it at the end (buffering all).
        self.yieldable.sort(key=(lambda x: x[0].datetime))


IpPacket.register_subtype(SipPacket)

# # Simple test/example
# import datetime
# sip_packet = IpPacket.create(
#     datetime.datetime.now(),
#     'UDP',
#     ('1.2.3.4', 1234),
#     ('1.2.3.4', 1234),
#     'INVITE sip:+123@1.2.3.4...\r\nCSeq: 667 INVITE\r\n'
# )
# assert isinstance(sip_packet, SipPacket), \
#     'Packet is of type: %r' % (type(sip_packet),)
# assert sip_packet.cseq == ('667', 'INVITE'), \
#     'CSeq is: %r' % (sip_packet.cseq,)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine Data source lib
# Copyright (C) 2011-2015,2018 Walter Doekes, OSSO B.V.

import datetime
import os
import socket
import struct
import sys

try:
    import pcap  # python-libpcap
except ImportError:
    try:
        import pcapy as pcap
    except ImportError:
        if sys.version_info.major < 3:
            raise ImportError('Please apt install python-libpcap (pylibpcap)')
        raise ImportError('Please pip install pcapy (libpcap replacement)')

from .libfile import extract_to_tempfile
from .libproto import IpPacket

if sys.version_info.major < 3:
    tobytes = bytearray  # mutable byte array, behaving like py3-bytes
else:
    tobytes = (lambda x: x)


class Skipping(BaseException):
    pass


class PcapReader(object):
    '''
    Reads a pcap file.
    '''
    protocols = {
        socket.IPPROTO_TCP: 'TCP',
        socket.IPPROTO_UDP: 'UDP',
        socket.IPPROTO_ICMP: 'ICMP'
    }

    def __init__(self, filenames, pcap_filter=None, min_date=None,
                 max_date=None):
        if not pcap:
            raise ImportError('PcapReader requires pylibpcap support '
                              '(python-libpcap)')

        self.filenames = filenames
        self.pcap_filter = pcap_filter

        # This way we'll only need to watch out for a single exception when
        # iterating.
        class BogoPcap():
            def next(self):
                raise TypeError("'NoneType' object is not iterable")
        self.pcap = BogoPcap()

        self.min_date = min_date
        self.max_date = max_date

        self.filename = None
        self.link_type = None

        self.warnings = set()  # store what we've have warned about

    def __iter__(self):
        return self

    def _get_next_packet(self):
        while True:
            # Re-open a new file until we have a packet
            try:
                # Fetch the next entry or raise TypeError:
                # "'NoneType' object is not iterable"
                next_packet = self.pcap.next()

                if len(next_packet) == 3:
                    # pylibpcap
                    (pktlen, data, timestamp) = next_packet
                else:
                    # pcapy
                    (pkthdr, data) = next_packet
                    # pcapy returns (None, '') for the last packet
                    if pkthdr is None:
                        raise TypeError('EOF')
                    timestamp = pkthdr.getts()
                    timestamp = timestamp[0] + timestamp[1] * 0.000001

            except TypeError:
                self._open_next_file()
            else:
                break

        # On python2, convert to list-of-integers.
        data = tobytes(data)

        return data, timestamp

    def _open_next_file(self):
        # In the past, python-libpcap did not clean up its fds. In
        # 0.6.4-1 it does though and the following is a no-op.
        self.pcap = None

        # Are we done?
        if not self.filenames:
            raise StopIteration()

        # Next file!
        self.filename = self.filenames.pop(0)
        # There is no pcap_fopen_offline(3pcap) call, so we'll
        # have to use a temp-file if the file was zipped.
        filename, is_tempfile = extract_to_tempfile(self.filename)

        try:
            self.pcap = pcap.pcapObject
        except AttributeError:
            # pcapy
            self.pcap = pcap.open_offline(filename)
        else:
            # pylibpcap
            self.pcap = pcap.pcapObject()
            self.pcap.open_offline(filename)

        if is_tempfile:
            os.unlink(filename)

        if self.pcap_filter:
            self.pcap.setfilter(self.pcap_filter, 0, 0)

        # Set link type, needed below
        self.link_type = self.pcap.datalink()

    def _get_frame_payload(self, data):
        # Get frame payload (pcap-linktype(7))
        # http://www.tcpdump.org/linktypes.html
        if self.link_type == pcap.DLT_RAW:  # Don't know??
            payload = data
        elif self.link_type == pcap.DLT_EN10MB:  # 0x1 Ethernet
            # to_mac = data[0:6]
            # from_mac = data[6:12]
            payload = data[12:]
        elif self.link_type == pcap.DLT_LINUX_SLL:  # 0x71 Linux Cooked SSL
            # packet_type = data[0:2]
            # arphdr_type = data[2:4]
            # lladdr_len = data[4:6] # =6 for mac
            # lladdr = data[6:14] # first 6 bytes for macaddr
            payload = data[14:]
        else:
            raise NotImplementedError(
                'Not implemented link type %d (0x%x) in %s' %
                (self.link_type, self.link_type, self.filename))

        return payload

    def _get_ethernet_data(self, payload):
        # Get ethernet data
        # http://en.wikipedia.org/wiki/EtherType
        data = None
        while True:
            if self.link_type == pcap.DLT_RAW:
                data = payload
                break
            elif payload[0:2] == b'\x08\x00':    # IPv4
                data = payload[2:]
                break
            elif payload[0:2] == b'\x81\x00':    # 802.1Q
                # tci = payload[2:2]  # pcp+cfi+vid
                payload = payload[4:]
                continue
            elif payload[0:2] == b'\x88\xa8':    # 802.1ad (Q-in-Q)
                raise NotImplementedError('VLAN-tagged ethernet frame '
                                          'decoding not implemented yet.')
            elif payload[0:2] == b'\x91\x00':    # 802.1QinQ (non-standard)
                raise NotImplementedError('VLAN-tagged ethernet frame '
                                          'decoding not implemented yet.')
            elif payload[0:2] == b'\x86\xdd':    # IPv6
                break  # ignore
            else:                                # Other stuff (like ARP)
                break  # ignore

        if data is None:
            raise Skipping()

        return data

    def _decode_ipv4(self, data, timestamp):
        version = data[0] >> 4
        header_len = (data[0] & 0x0f) << 2
        if version != 4:
            raise ValueError('How did you get a version %d in an IPv4 '
                             'header?' % (version,))

        flags = data[6] >> 5
        fragment_offset = struct.unpack('>H', data[6:8])[0] & 0x1fff
        if flags & 2:
            msg = ('(packet defragmentation on t %f not implemented yet, '
                   'suppressing warning)')
            self.warn_once('more_fragments', msg % (timestamp,))
            # but, carry on
        if fragment_offset:
            msg = '(skipping IP fragment on t %f, suppressing warning)'
            self.warn_once('fragment', msg % (timestamp,))
            raise Skipping()

        try:
            proto_num = data[9]
            ip_proto = self.protocols[proto_num]
        except KeyError:
            msg = ('(skipping unknown IP protocol %d on t %f, suppressing '
                   'warning)' % (proto_num, timestamp))
            self.warn_once('proto:%d' % (proto_num,), msg)
            raise Skipping()

        from_ = socket.inet_ntoa(bytes(data[12:16]))
        to = socket.inet_ntoa(bytes(data[16:20]))

        # IP => TCP/UDP/ICMP/fragment
        body = data[header_len:]

        return from_, to, ip_proto, body

    def __next__(self):
        while True:
            try:
                data, timestamp = self._get_next_packet()

                if self.min_date and self.min_date > timestamp:
                    continue
                if self.max_date and self.max_date < timestamp:
                    continue

                data = self._get_frame_payload(data)
                data = self._get_ethernet_data(data)
                from_, to, ip_proto, data = self._decode_ipv4(data, timestamp)
            except Skipping:
                continue

            if ip_proto in ('TCP', 'UDP'):
                from_ = (from_, struct.unpack('>H', data[0:2])[0])  # add port
                to = (to, struct.unpack('>H', data[2:4])[0])        # add port

                if ip_proto == 'TCP':
                    # FIXME: we need another layer for reassembling TCP
                    # into a stream before attempting to do app-protocol
                    # decoding on it.
                    data_offset = (data[12] >> 4) * 4
                    data = data[data_offset:]

                else:
                    data = data[8:]

            elif ip_proto == 'ICMP':
                # Parsing ICMP is nice if we want to trace port-
                # unreachable messages.
                msg = ('(skipping IP ICMP protocol on t %f, not yet '
                       'implemented, suppressing warning)')
                self.warn_once('proto:icmp', msg % (timestamp,))
                continue
            else:
                raise NotImplementedError(ip_proto)

            datetime_ = datetime.datetime.fromtimestamp(timestamp)
            return IpPacket.create(datetime_, ip_proto, from_, to, data)
    next = __next__

    def warn_once(self, key, message):
        if key not in self.warnings:
            sys.stderr.write(message + '\n')
            self.warnings.add(key)



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine main (SIP Examine)
# Copyright (C) 2011-2015 Walter Doekes, OSSO B.V.

from datetime import datetime, timedelta
from time import mktime, strptime
import re
import sys

from . import sipzamine
from .argparse14191 import ArgumentParser14191
from .libdata import PcapReader
from .libprotosip import SipDialogs

# Matches dialogs and times and looks for certain info.
__author__ = 'Walter Doekes'
__copyright__ = 'Copyright (C) Walter Doekes, OSSO B.V. 2011-2015,2018'
__licence__ = 'GPLv3+'
__version__ = '0.2'

try:
    cmp
except NameError:
    cmp = None  # FIXME: python3


class epochtime_without_date(object):
    """
    Magic float that ignores the date so time-only comparisons can be
    done.  Take care to ensure that in datetime comparisons, this object
    is on the LHS, so these overloaded __cmp__ operator can do its job.

    NOTE: We calculate the delta between daylight saving time in the initial
    value and the compared value ONLY ONCE.  This will produce crazy results
    around DST changes.  Also, around hour 00:00 we cannot guarantee any
    meaningful result either.  This is a quick hack that works in 95% of the
    cases.
    """
    def __init__(self, floatval):
        sys.stderr.write(
            '(warning: partial date parsing is inaccurate! use only if '
            'you know understand the limitations)\n')
        self.floatval = floatval
        self.dstdelta = None

    def __float__(self):
        return self.floatval

    def __cmp__(self, other):
        otherval = float(other)

        if self.dstdelta is None:
            self.dstdelta = (
                (datetime.fromtimestamp(otherval) -
                 datetime.utcfromtimestamp(otherval)).seconds -
                (datetime.fromtimestamp(self.floatval) -
                 datetime.utcfromtimestamp(self.floatval)).seconds
            )

        selfval = self.floatval % 86400.0
        otherval = (otherval + self.dstdelta) % 86400.0
        return cmp(selfval, otherval)


def my_regex(regexstring):
    try:
        return re.compile(regexstring)
    except Exception:
        raise ValueError()


def my_strptime(timestring):
    for style, prefix in (
            ('%Y-%m-%d %H:%M:%S', ''),
            ('%Y-%m-%d %H:%M', ''),
            ('%Y-%m-%d', ''),
            ('%Y-%m-%d %H:%M:%S', '2000-01-01 '),
            ('%Y-%m-%d %H:%M', '2000-01-01 ')):
        try:
            parsed = strptime(prefix + timestring, style)
        except ValueError:
            pass
        else:
            ret = mktime(parsed)
            if prefix:
                ret = epochtime_without_date(ret)
            break
    else:
        raise ValueError('Invalid time format, use YYYY-MM-DD '
                         'HH:MM:SS or shortened')
    return ret


def my_timedelta(floatstring):
    if '.' in floatstring:
        # Add enough zeroes so the int value is large enough..
        floatstring = '0%s000000' % (floatstring,)
        num = [int(floatstring.split('.', 1)[0]),       # seconds
               int(floatstring.split('.', 1)[1][0:6])]  # milliseconds
    else:
        num = [int(floatstring),  # seconds
               0]                 # milliseconds
    if sum(num) == 0:
        raise ValueError('Specifying a zero time breaks boolean tests')
    return timedelta(seconds=num[0], milliseconds=num[1])


def parse_args():
    # Example: sipzamine -m '^INVITE' -H 'm=audio ([0-9]+)' \
    #                       -p 'host 1.2.3.4' 5060.pcap.00
    description = 'Search and examine SIP transactions/dialogs'
    parser = ArgumentParser14191(prog='sipzamine', description=description)

    parser.add_argument(
        'files', metavar='PCAP', nargs='+',
        help='pcap files to parse')
    parser.add_argument(
        '--pcap', '-p', metavar='filter',
        help='pcap filter expression')

    # FIXME: remark that the searches are performed on the header lines and
    # can be anchored as such
    parser.add_argument(
        '--pmatch', '-m', metavar='regex', action='append', type=my_regex,
        help=('any packet in dialog must match regex (can be used '
              'multiple times), e.g. ^INVITE to match calls'))
    # FIXME: we may need to tweak the --option-name here too, and the
    # description
    parser.add_argument(
        '--amatch', '-M', metavar='regex', action='append', type=my_regex,
        help='all packets in dialog must match regex (can be used '
             'multiple times), e.g. ^(SIP/2.0|INVITE|BYE) to match calls '
             'without an ACK')
    parser.add_argument(
        '--highlight', '-H', metavar='regex', action='append', type=my_regex,
        help=('highlight first matchgroup in packets (multiple '
              'highlights are identified by letters a..z)'))

    parser.add_argument(
        '--dateskew', metavar='seconds', default=0, type=int,
        help=('offset added to all dates, can be negative (use when pcap '
              'clock was off)'))

    parser.add_argument(
        '--mindate', metavar='date', type=my_strptime,
        help='packets must be younger than specified date')
    parser.add_argument(
        '--maxdate', metavar='date', type=my_strptime,
        help='packets must be older than specified date')

    parser.add_argument(
        '--mindur', metavar='seconds', type=my_timedelta,
        help='dialogs/transactions must be shorter than duration')
    parser.add_argument(
        '--maxdur', metavar='seconds', type=my_timedelta,
        help='dialogs/transactions must be longer than duration')

    parser.add_argument(
        '--retransmits', metavar='count', default=0, type=int,
        help='at least count retransmits must be involved')

    parser.add_argument(
        '--contents', action='store_true', default=False,
        help='show complete packet contents')

    return parser.parse_args()


def adjust_times(args):
    """
    Update the search dates according to the date skew
    """
    if args.dateskew:
        args.dateskew = timedelta(seconds=args.dateskew)
        if args.mindate:
            args.mindate += args.dateskew
        if args.maxdate:
            args.maxdate += args.dateskew


def add_dialog_filters(sipdialogs, args):
    """
    Optionally add duration and search filters (try to put the light
    weight ones first)
    """
    if args.mindur:
        sipdialogs = sipzamine.minduration_filter(
            sipdialogs, min_duration=args.mindur)

    if args.maxdur:
        sipdialogs = sipzamine.maxduration_filter(
            sipdialogs, max_duration=args.maxdur)

    if args.amatch:
        for amatch in args.amatch:
            sipdialogs = sipzamine.allheaders_filter(
                    sipdialogs, header_match=amatch)
    if args.pmatch:
        for pmatch in args.pmatch:
            sipdialogs = sipzamine.anyheader_filter(
                sipdialogs, header_match=pmatch)

    if args.retransmits:
        sipdialogs = sipzamine.retransmits_filter(
            sipdialogs, count=args.retransmits)

    return sipdialogs


def main():
    args = parse_args()
    adjust_times(args)

    # Create a packet reader
    reader = PcapReader(
        args.files, pcap_filter=args.pcap,
        min_date=args.mindate, max_date=args.maxdate)

    # Optionally add a date skew on the packets
    if args.dateskew:
        reader = sipzamine.dateskew_filter(reader, skew=args.dateskew)

    # Convert the packets into SIP dialogs
    reader = SipDialogs(reader)

    # Add filters
    reader = add_dialog_filters(reader, args)

    # Call main with our pimped reader
    sipzamine.main(
        reader, packet_highlights=args.highlight,
        show_contents=args.contents)


if __name__ == '__main__':
    main()

# Example usage:
#
# $ sipzamine -m 'sip:\+315' -H '^BYE' --pcap 'host banana' \
#                stored.pcap
# (or)
# $ /usr/sbin/tcpdump -nnvs0 -r stored.pcap host banana |
#       sipzamine -m 'sip:\+315' -H '^BYE' -
#
# Example output:
#
# [ 179978155f707e3622c0886752336210@22.22.22.22 ]
# 2011-11-23 22:27:20.746782 apple:5060 > banana:5060 102 INVITE
# 2011-11-23 22:27:20.747508 banana:5060 > apple:5060 102 INVITE(100)
# 2011-11-23 22:27:20.783424 banana:5060 > apple:5060 102 INVITE(200)
# 2011-11-23 22:27:20.783956 apple:5060 > banana:5060 102 ACK
# 2011-11-23 22:27:41.665581 apple:5060 > banana:5060 103 BYE <--
# 2011-11-23 22:27:41.665721 banana:5060 > apple:5060 103 BYE(200)
#
# [ 64e9278b4cdabb7c02f8c54f301937e7@apple ]
# 2011-11-23 22:28:16.875647 apple:5060 > banana:5060 102 INVITE
# 2011-11-23 22:28:16.876433 banana:5060 > apple:5060 102 INVITE(100)
# 2011-11-23 22:28:16.901755 banana:5060 > apple:5060 102 INVITE(200)
# 2011-11-23 22:28:16.902327 apple:5060 > banana:5060 102 ACK
# 2011-11-23 22:28:24.363193 apple:5060 > banana:5060 103 BYE <--
# 2011-11-23 22:28:24.363352 banana:5060 > apple:5060 103 BYE(200)



#!/usr/bin/env python
from distutils.core import setup

with open('README.rst') as file:
    long_description = file.read()


setup(
    name='sipzamine',
    version='0.4',
    entry_points={'console_scripts': ['sipzamine = sipzamine.__main__:main']},
    packages=['sipzamine'],
    data_files=[('share/doc/sipzamine', [
        'LICENSE.txt', 'README.rst', 'sipzamine/argparse-LICENSE.txt'])],
    description='SIP dialog examine utility (formerly sipcaparseye)',
    long_description=long_description,
    author='Walter Doekes, OSSO B.V.',
    author_email='wjdoekes+sipzamine@osso.nl',
    url='https://github.com/ossobv/sipzamine',
    license='GPLv3+',
    platforms=['linux'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        ('License :: OSI Approved :: GNU General Public License v3 '
         'or later (GPLv3+)'),
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Communications :: Telephony',
        'Topic :: System :: Networking :: Monitoring',
    ],
    install_requires=[
        # 'pylibpcap' || 'pcapy'
        # The former refuses to install easily using pip, use your
        # OS package manager if possible.
    ],
)

# vim: set ts=8 sw=4 sts=4 et ai tw=79:
